SystemOrganization addCategory: #'VMConstruction-Plugins-AioPlugin'!

InterpreterPlugin subclass: #AioPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-AioPlugin'!

!AioPlugin commentStamp: '<historical>' prior: 0!
AioPlugin provides a primitive interface to asynchronous input/output events. Different
classes are required for various operating system platforms. All implementations are expected
to use the module name "AioPlugin" regardless of their class name. This provides a uniform
primitive interface for asynchronous events on all supported platforms.

Aio events are externally occuring events such as "data is now available on an IO channel"
or "an IO channel is now available for writing". These events occur asynchronously with
respect to the Squeak VM. AioPlugin arranges for event forwarding such that each event
causes a corresponding Smalltalk semaphore to be signalled.

To create an event handler in Squeak, register an external semaphore, establish event
forwarding to the semaphore, and start a Smalltalk process to wait on the semaphore. The
process should loop on the semaphore wait, taking appropriate action when the event is
signalled, and terminating when the event forwarding is disabled.

The supported primitives are:
  primitiveAioEnable - Enable asynchronous notification for a descriptor.
  primitiveAioHandle - Handle asynchronous event notification for a descriptor.
  primitiveAioSuspend - Temporarily suspend asynchronous event notification for a descriptor.
  primitiveAioDisable - Definitively disable asynchronous event notification for a descriptor.
  primitiveOSFileHandle - Answer the handle (Unix file number) for a FileStream.
  primitive OSSocketHandle - Answer the handle (Unix file number) for a Socket.

Currently the Unix VM supports aio event notification, and class UnixAioPlugin provides
the required primitive support.
!

!AioPlugin class methodsFor: 'translation' stamp: 'dtl 5/1/2012 19:51'!
declareCVarsIn: cg

	"Additional declarations are done in my subclasses. This permits a meaningful
	module identification string to be written with the name of the actual subclass
	being generated."

	cg addHeaderFile: '"FilePlugin.h"'.
	cg addHeaderFile: '"SocketPlugin.h"'.
	cg addHeaderFile: '"config.h"
#define SESSIONIDENTIFIERTYPE int'.	"Data type for sessionID, should generate compiler warnings
									if someone changes the data type elsewhere in the VM."
	cg addHeaderFile: '"sqaio.h"'.
! !

!AioPlugin class methodsFor: 'version identification' stamp: 'dtl 9/1/2003 17:23'!
installedModuleName
	"Answer the module name string for the plugin which is actually running in the VM"

	"self installedModuleName"

	<primitive: 'primitiveModuleName' module: 'AioPlugin'>
	^ nil
! !

!AioPlugin class methodsFor: 'version identification' stamp: 'dtl 9/1/2003 17:22'!
installedModuleVersionString
	"Answer the version string for the plugin which is actually running in the VM"

	"self installedModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'AioPlugin'>
	^ nil
! !

!AioPlugin class methodsFor: 'translation' stamp: 'dtl 7/4/2003 19:17'!
moduleName

	^ 'AioPlugin'! !

!AioPlugin class methodsFor: 'translation' stamp: 'dtl 11/28/2010 10:25'!
platformName 
	"After Squeak version 3.6, #platformName was moved to SmalltalkImage Some
	versions of Pharo move this to OSPlatform and issue deprecation warnings
	about the other usages."

	"self platformName"

	^ (((Smalltalk hasClassNamed: #OSPlatform)
			and: [(Smalltalk at: #OSPlatform)
					respondsTo: #platformName])
		ifTrue: [Smalltalk at: #OSPlatform]
		ifFalse: [((Smalltalk classNamed: 'SmalltalkImage')
				ifNil: [^ Smalltalk osVersion]) current]) platformName! !

!AioPlugin class methodsFor: 'translation' stamp: 'dtl 7/26/2010 22:50'!
shouldBeTranslated

	^ false
! !

!AioPlugin class methodsFor: 'version identification' stamp: 'dtl 5/1/2012 19:45'!
versionString
	"Important: When this method is changed, the instance side method must also be
	changed to match."

	"self versionString"

	"2.0 supports 64bit code base"

	^ self new versionString
! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 4/29/2006 11:13'!
fileDescriptorFrom: aSQFileByteArray
	"Answer the OS file descriptor, an integer value, from a SQFile data structure
	byte array, or answer -1 if unable to obtain the file descriptor (probably due
	to receiving an incorrect type of object as aFileHandle)."

	"return type should be int, but skip the declaration to permit inlining"
	(self isSQFileObject: aSQFileByteArray) ifFalse: [^ -1].
	^ self fileno: (self fileHandleFrom: aSQFileByteArray)
! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 10:41'!
fileHandleFrom: sqFileStructByteArray
	"Answer a file handle from a SQFile structure. On most platforms, this
	will be a (FILE *). On Win32, it is a HANDLE."

	| sqFile |
	<returnTypeC: 'FILEHANDLETYPE'>
	<var: 'sqFile' declareC: 'SQFile *sqFile'>
	sqFile := interpreterProxy arrayValueOf: sqFileStructByteArray.
	^ self cCode: 'sqFile->file' inSmalltalk: [0]
! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'eem 11/19/2014 22:32'!
fileRecordSize
	"Answer the size of a SQFile data structure in bytes."

	^ self sizeof: #SQFile! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 11:05'!
fileValueOf: anSQFileRecord
	"Return a pointer to the first byte of of the SQFile data structure file record within
	anSQFileRecord, which is expected to be a ByteArray of size self>>fileRecordSize."

	<returnTypeC: 'SQFile *'>
	^ interpreterProxy arrayValueOf: anSQFileRecord
! !

!AioPlugin methodsFor: 'private' stamp: 'dtl 4/27/2006 20:13'!
getThisSessionIdentifier

	^ interpreterProxy getThisSessionID
! !

!AioPlugin methodsFor: 'initialize-release' stamp: 'dtl (auto pragmas dtl 2010-11-28) 12/26/2007 09:37'!
initialiseModule

	<export: true>
	^ true
! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'eem 4/17/2013 17:23'!
isNonNullSQFile: objectPointer
	"Check for the common failure mode of a SQFile record with all zeros."

	| idx sqFileBytes  |
	<var: 'sqFileBytes' type: #'unsigned char *'>
	<var: 'idx' type: #unsigned>
	sqFileBytes := interpreterProxy arrayValueOf: objectPointer.
	idx := 0.
	[idx < self fileRecordSize]
		whileTrue:
			[(sqFileBytes at: idx) ~= 0 ifTrue: [^ true].
			idx := idx + 1].
	^ false
! !

!AioPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 12:03'!
isNullSQSocket: objectPointer
	"Check for the common failure mode of a SQSocket record with all zeros."

	| idx sqSocketBytes |
	<var: 'sqSocketBytes' type: 'unsigned char *'>
	sqSocketBytes := interpreterProxy arrayValueOf: objectPointer.
	idx := 0.
	[idx < self socketRecordSize]
		whileTrue:
			[((sqSocketBytes at: idx) ~= 0) ifTrue: [^ false].
			idx := idx + 1].
	^ true
! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 4/29/2006 12:45'!
isSQFileObject: objectPointer 
	"Answer true if objectPointer appears to be a valid SQFile ByteArray.
	This check is appropriate if objectPointer has been passed as a parameter
	to a primitive, and is expected to represent a valid file reference."

	^ (((interpreterProxy isBytes: objectPointer)
		and: [(interpreterProxy byteSizeOf: objectPointer) = self fileRecordSize])
		and: [self isValidFileSession: objectPointer])
		and: [self isNonNullSQFile: objectPointer]! !

!AioPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl 4/29/2006 11:54'!
isSQSocketObject: objectPointer
	"Answer true if objectPointer appears to be a valid SQSocket ByteArray. This check
	is appropriate if objectPointer has been passed as a parameter to a primitive, and
	is expected to represent a valid socket reference."

	^ (((interpreterProxy isBytes: objectPointer)
		and: [(interpreterProxy byteSizeOf: objectPointer) = self socketRecordSize]))
		and: [(self isNullSQSocket: objectPointer) not]
! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'dtl 4/29/2006 11:25'!
isValidFileSession: objectPointer
	"Answer true if the file session matches the current interpreter session identifier."

	 ^ self getThisSessionIdentifier =
		(self sessionIdentifierFromSqFile: (interpreterProxy arrayValueOf: objectPointer))
! !

!AioPlugin methodsFor: 'initialize-release' stamp: 'dtl (auto pragmas dtl 2010-11-28) 6/7/2008 14:53'!
moduleUnloaded: aModuleName
	"The module with the given name was just unloaded.
	Make sure we have no dangling references."
	<export: true>
	<var: #aModuleName type: 'char *'>
! !

!AioPlugin methodsFor: 'primitives' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 17:50'!
primitiveModuleName
	"Answer a string containing the module name string for this plugin."

	<export: true>
	interpreterProxy pop: 1 thenPush: (self stringFromCString: self getModuleName)
! !

!AioPlugin methodsFor: 'primitives' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/1/2003 20:11'!
primitiveVersionString
	"Answer a string containing the version string for this plugin."

	<export: true>
	interpreterProxy pop: 1 thenPush: (self stringFromCString: self versionString)
! !

!AioPlugin methodsFor: 'private - SQFile access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 11:23'!
sessionIdentifierFromSqFile: sqFile
	"Answer the session identifier from a SQFile structure. For a valid file
	reference, this identifier will match the session identifier supplied by
	the interpreter."

	<inline: true>
	<returnTypeC: 'SESSIONIDENTIFIERTYPE'>
	<var: 'sqFile' type: 'SQFile *'>
	^ self cCode: 'sqFile->sessionID' inSmalltalk: [-1]
! !

!AioPlugin methodsFor: 'initialize-release' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/1/2003 19:55'!
shutdownModule

	<export: true>
! !

!AioPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 13:13'!
socketDescriptorFrom: sqSocketOop
	"Answer the OS file descriptor, an integer value, from a SQSocket data structure,
	or answer -1 if unable to obtain the file descriptor (probably due to receiving
	an incorrect type of object as aFileHandle).

	Warning: The first element of privateSocketStruct happens to be the Unix file
	number of the socket. See sqUnixSocket.c for the definition. This method takes
	advantage of this, and will break if anyone ever redefines the data structure."

	| sqSocket privateSocketStruct |
	<returnTypeC: 'int'>
	<var: 'sqSocket' declareC: 'SQSocket *sqSocket'>
	<var: 'privateSocketStruct' declareC: 'void *privateSocketStruct'>
	sqSocket := self socketValueOf: sqSocketOop.
	privateSocketStruct := self cCode: 'sqSocket->privateSocketPtr'.
	privateSocketStruct = 0
		ifTrue: [^ -1].	"Something is wrong, bail out before we core dump."
	^ self cCode: '* (int *) privateSocketStruct' inSmalltalk: [-1].
! !

!AioPlugin methodsFor: 'private - SQSocket access' stamp: 'eem 3/22/2015 19:46'!
socketRecordSize
	"Answer the size of a SQSocket data structure in bytes."
	<inline: true>
	^ self sizeof: #SQSocket! !

!AioPlugin methodsFor: 'private - SQSocket access' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 11:47'!
socketValueOf: anSQSocketRecord
	"Return a pointer to the first byte of of the SQsocket data structure socket record within
	anSQSocketRecord, which is expected to be a ByteArray of size self>>socketRecordSize."

	<returnTypeC: 'SQSocket *'>
	^ interpreterProxy arrayValueOf: anSQSocketRecord
! !

!AioPlugin methodsFor: 'private' stamp: 'dtl (auto pragmas dtl 2010-11-28) 4/29/2006 18:40'!
stringFromCString: aCString
	"Answer a new String copied from a null-terminated C string.
	Caution: This may invoke the garbage collector."

	| len newString |
	<var: 'aCString' type: 'const char *'>
	len := self strlen: aCString.
	newString := interpreterProxy
		instantiateClass: interpreterProxy classString
		indexableSize: len.
	self st: (interpreterProxy arrayValueOf: newString)
		rn: aCString
		cpy: len. "(char *)strncpy()"
	^ newString

! !

!AioPlugin methodsFor: 'version identification' stamp: 'dtl 5/1/2012 19:44'!
versionString
	"Answer a string containing the version string for this plugin. Handle MNU
	errors, which can occur if class InterpreterPlugin has been removed from
	the system.

	Important: When this method is changed, the class side method must also be
	changed to match."

	"2.0 supports 64bit code base"

	| version |
	<returnTypeC: 'char *'>
	<var: 'version' declareC: 'static char version[]= "2.2.6"'>
	^ self cCode: 'version' inSmalltalk: ['2.2.6']

! !

AioPlugin subclass: #UnixAioPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMConstruction-Plugins-AioPlugin'!

!UnixAioPlugin commentStamp: '<historical>' prior: 0!
UnixAioPlugin implements the Unix platform methods for AioPlugin.!

!UnixAioPlugin class methodsFor: 'translation' stamp: 'dtl 4/29/2006 10:48'!
declareCVarsIn: cg
	"Define FILEHANDLETYPE. The #include <unistd.h> is completely unnecessary, but
	serves as a way to kludge the #define into the header."

	cg addHeaderFile: '<unistd.h>
#define FILEHANDLETYPE FILE *  /* the type of low level stream to be used in a struct SQFile */'
! !

!UnixAioPlugin class methodsFor: 'translation' stamp: 'dtl 7/26/2010 22:51'!
isSuitablePluginForPlatform: platformName
	"Answer true is this is an instance of the class which is responsible for
	representing the OS process for the Squeak VM running on platform.
	Newer versions of VMMaker use this mechanism"

	^ platformName = 'unix'
		or: [platformName = 'Mac OS']! !

!UnixAioPlugin class methodsFor: 'translation' stamp: 'dtl 7/26/2010 22:51'!
shouldBeTranslated

	^ true
! !

!UnixAioPlugin methodsFor: 'private - event forwarding' stamp: 'dtl (auto pragmas dtl 2010-11-28) 9/13/2005 22:03'!
aioForward: fd withData: data andFlags: flags
	"This function is called to signal a Smalltalk Semaphore when an asynchronous event is
	detected. When translated to C, the name of this method is aioForwardwithDataandFlags.
	The event handler is set up by #primitiveAioHandle."

	| semaIndex pfd |
	<returnTypeC: 'void'>
	<var: 'fd' type: 'int'>
	<var: 'data' type: 'void *'>
	<var: 'flags' type: 'int'>
	<var: 'pfd' declareC: 'int *pfd'>
	pfd := data.
	semaIndex := self cCode: '*pfd' inSmalltalk: [0].
	interpreterProxy signalSemaphoreWithIndex: semaIndex
! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl (auto pragmas dtl 2010-11-28) 12/19/2007 18:54'!
primitiveAioDisable
	"Definitively disable asynchronous event notification for a descriptor. The
	parameter is an OS level integer file descriptor."

	| fd |
	<export: true>
	(interpreterProxy nilObject = (interpreterProxy stackValue: 0))
		ifTrue: [^ interpreterProxy primitiveFail].
	fd := interpreterProxy stackIntegerValue: 0.
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	self aioDisable: fd.
	interpreterProxy pop: 2; pushInteger: fd
! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl (auto pragmas dtl 2010-11-28) 12/19/2007 18:56'!
primitiveAioEnable
	"Enable asynchronous notification for a descriptor. The first parameter is an OS
	level integer file descriptor. The second parameter is the index of a Semaphore to
	be notified, and the third parameter is a flag indicating that descriptor represents
	an external object and should not be closed on termination of aio handling. Answer
	the semaphore index."

	| fd semaIndex flags externalObject eventSemaphoreIndices |
	<export: true>
	<var: 'eventSemaphoreIndices' declareC: 'static int eventSemaphoreIndices[FD_SETSIZE]'>
	(interpreterProxy nilObject = (interpreterProxy stackValue: 2))
		ifTrue: [^ interpreterProxy primitiveFail].
	fd := interpreterProxy stackIntegerValue: 2.
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	semaIndex := interpreterProxy stackIntegerValue: 1.
	"The following is to prevent the need to coerce integers into pointers in clientData".
	eventSemaphoreIndices at: semaIndex put: semaIndex.
	externalObject := interpreterProxy stackObjectValue: 0.
	(externalObject = interpreterProxy trueObject)
		ifTrue: [flags := self cCode: 'AIO_EXT' inSmalltalk: [16]]
		ifFalse: [flags := 0].
	self cCode: 'aioEnable(fd, &(eventSemaphoreIndices[semaIndex]), flags)'.
	interpreterProxy pop: 4; pushInteger: semaIndex.! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'eem 12/12/2014 16:55'!
primitiveAioHandle
	"Handle asynchronous event notification for a descriptor. The first parameter is
	an OS level integer file descriptor. The remaining three parameters are Boolean
	flags representing the types of events for which notification is being requested:
	handle exceptions, handle for read, and handle for write.
	Flags are defined in the aio.h source as:
		AIO_X	(1<<0)	handle for exceptions
		AIO_R	(1<<1)	handle for read
		AIO_W	(1<<2)	handle for write"

	| fd exceptionWatch readWatch writeWatch flags |
	<export: true>
	(interpreterProxy nilObject = (interpreterProxy stackValue: 3))
		ifTrue: [^ interpreterProxy primitiveFail].
	fd := interpreterProxy stackIntegerValue: 3.
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	exceptionWatch := interpreterProxy stackObjectValue: 2.
	readWatch := interpreterProxy stackObjectValue: 1.
	writeWatch := interpreterProxy stackObjectValue: 0.
	flags := 0.
	(exceptionWatch = interpreterProxy trueObject)
		ifTrue: [flags := flags bitOr: (self cCode: 'AIO_X' inSmalltalk: [1])].
	(readWatch = interpreterProxy trueObject)
		ifTrue: [flags := flags bitOr: (self cCode: 'AIO_R' inSmalltalk: [2])].
	(writeWatch = interpreterProxy trueObject)
		ifTrue: [flags := flags bitOr: (self cCode: 'AIO_W' inSmalltalk: [4])].
	self aio: fd Han: #aioForward:withData:andFlags: dle: flags.
	interpreterProxy pop: 5; pushInteger: flags
! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl (auto pragmas dtl 2010-11-28) 12/19/2007 18:57'!
primitiveAioSuspend
	"Temporarily suspend asynchronous event notification for a descriptor. The first
	parameter is an OS level integer file descriptor. The remaining three parameters
	are Boolean flags representing the types of events for which notification is being
	requested: handle exceptions, handle for read, and handle for write.
	Flags are defined in the aio.h source as:
		AIO_X	(1<<0)	handle for exceptions
		AIO_R	(1<<1)	handle for read
		AIO_W	(1<<2)	handle for write"

	| fd exceptionWatch readWatch writeWatch flags |
	<export: true>
	(interpreterProxy nilObject = (interpreterProxy stackValue: 3))
		ifTrue: [^ interpreterProxy primitiveFail].
	fd := interpreterProxy stackIntegerValue: 3.
	(fd < 0) ifTrue: [^ interpreterProxy primitiveFail].
	exceptionWatch := interpreterProxy stackObjectValue: 2.
	readWatch := interpreterProxy stackObjectValue: 1.
	writeWatch := interpreterProxy stackObjectValue: 0.
	flags := 0.
	(exceptionWatch == interpreterProxy trueObject)
		ifTrue: [flags := flags bitOr: (self cCode: 'AIO_X' inSmalltalk: [1])].
	(readWatch == interpreterProxy trueObject)
		ifTrue: [flags := flags bitOr: (self cCode: 'AIO_R' inSmalltalk: [2])].
	(writeWatch == interpreterProxy trueObject)
		ifTrue: [flags := flags bitOr: (self cCode: 'AIO_W' inSmalltalk: [4])].
	self cCode: 'aioSuspend(fd, flags)'.
	interpreterProxy pop: 5; pushInteger: flags
! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl (auto pragmas dtl 2010-11-28) 7/4/2003 19:18'!
primitiveOSFileHandle
	"Take a struct SQFile from the stack, and answer the value of its Unix file number."

	| sqFileOop fileNo |
	<export: true>
	sqFileOop := interpreterProxy stackValue: 0.
	(self isSQFileObject: sqFileOop) ifFalse: [^ interpreterProxy primitiveFail].
	fileNo := self fileDescriptorFrom: sqFileOop.
	interpreterProxy pop: 2; pushInteger: fileNo
! !

!UnixAioPlugin methodsFor: 'primitives' stamp: 'dtl (auto pragmas dtl 2010-11-28) 7/4/2003 19:18'!
primitiveOSSocketHandle
	"Take a struct SQSocket from the stack, and answer the value of its Unix file number."

	| sqSocketOop fileNo |
	<export: true>
	sqSocketOop := interpreterProxy stackValue: 0.
	(self isSQSocketObject: sqSocketOop) ifFalse: [^ interpreterProxy primitiveFail].
	fileNo := self socketDescriptorFrom: sqSocketOop.
	(fileNo < 0) ifTrue: [^ interpreterProxy primitiveFail].
	interpreterProxy pop: 2; pushInteger: fileNo
! !
